use std::{
    io::{self, SeekFrom},
    str::FromStr,
};

use eyre::bail;

use crate::{
    error,
    parse::{Header, Op},
    reader::BagReader,
    Error, Result,
};

#[derive(Clone, Copy, Debug)]
pub enum Compression {
    None,
    Bz2,
    Lz4,
}

impl FromStr for Compression {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "none" => Compression::None,
            "bz2" => Compression::Bz2,
            "lz4" => Compression::Lz4,
            _ => bail!(error::UnsupportedCompression(s.to_string())),
        })
    }
}

impl Compression {
    pub fn decompress<'a, R: io::Read + 'a>(self, read: R) -> Box<dyn io::Read + 'a> {
        match self {
            Compression::None => Box::new(read),
            Compression::Bz2 => todo!("bz2 decompression"),
            Compression::Lz4 => Box::new(lz4_flex::frame::FrameDecoder::new(read)),
        }
    }
}

#[derive(Debug)]
pub struct ChunkHeader {
    pub compression: Compression,
    pub uncompressed_size: u32,
}

impl ChunkHeader {
    pub fn read<R: BagReader>(reader: &mut R, pos: u64) -> Result<ChunkHeader> {
        reader.seek(SeekFrom::Start(pos))?;
        let header = reader.read_header_op(Op::Chunk)?;
        ChunkHeader::from_header(header)
    }

    pub fn from_header(header: Header) -> Result<Self> {
        Ok(ChunkHeader {
            compression: header.read_string(b"compression")?.parse()?,
            uncompressed_size: header.read_u32(b"size")?,
        })
    }
}
